package me.mrletsplay.videobase.library;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import me.mrletsplay.mrcore.json.JSONObject;
import me.mrletsplay.mrcore.json.JSONParseException;
import me.mrletsplay.mrcore.json.converter.JSONConverter;
import me.mrletsplay.mrcore.json.converter.SerializationOption;
import me.mrletsplay.videobase.VideoBase;
import me.mrletsplay.videobase.util.Hash;

public class Library {

	public static final String
		METADATA_NAME = "meta.json",
		VIDEO_METADATA_SUFFIX = ".meta.json";

	private Path path;
	private boolean readOnly;
	private List<Video> videos;

	public Library(Path path, boolean readOnly, List<Video> videos) {
		Objects.requireNonNull(path, "path");
		Objects.requireNonNull(videos, "videos");

		this.path = path;
		this.readOnly = readOnly;
		this.videos = videos;
	}

	public Path getPath() {
		return path;
	}

	public boolean isReadOnly() {
		return readOnly;
	}

	public List<Video> getVideos() {
		return videos;
	}

	public Video findVideoById(String id) {
		return videos.stream()
			.filter(v -> v.getId().equals(id))
			.findFirst().orElse(null);
	}

	/**
	 * Finds the videos for the given series and/or author sorted by their index
	 * @param series The series or {@code null} to not filter by series
	 * @param author The author or {@code null} to not filter by author
	 * @return The videos matching the given criteria
	 */
	public List<Video> findVideosBy(String series, String author) { // TODO: Take VideoMetadata to allow any parameter from meta
		return videos.stream()
			.filter(v -> (series == null || series.equals(v.getMetadata().getSeries()))
				&& (author == null || author.equals(v.getMetadata().getAuthor())))
			.sorted(Comparator.comparing(v -> v.getMetadata().getIndex()))
			.collect(Collectors.toList());
	}

	public List<String> getSeries() {
		return videos.stream()
			.map(v -> v.getMetadata().getSeries())
			.distinct()
			.collect(Collectors.toList());
	}

	public List<String> getAuthors() {
		return videos.stream()
			.map(v -> v.getMetadata().getAuthor())
			.distinct()
			.collect(Collectors.toList());
	}

	public boolean updateMetadata(Video video, VideoMetadata metadata) {
		if(readOnly) return false;

		VideoMetadata oldMetadata = video.getMetadata();
		VideoMetadata newMetadata = VideoMetadata.inherit(oldMetadata.getParent(), metadata);

		Path videoPath = video.getPath().toAbsolutePath();
		try {
			Files.writeString(videoPath.getParent().resolve(videoPath.getFileName().toString() + VIDEO_METADATA_SUFFIX), newMetadata.toJSON(SerializationOption.DONT_INCLUDE_CLASS).toFancyString(), StandardCharsets.UTF_8);
			video.setMetadata(newMetadata);
			return true;
		} catch (IOException e) {
			return false;
		}
	}

	public boolean updateMetadata(String videoId, VideoMetadata metadata) {
		Video video = findVideoById(videoId);
		if(video == null) return false;
		return updateMetadata(video, metadata);
	}

	public static Library load(Path path, boolean readOnly) {
		List<Video> videos = new ArrayList<>();
		load(path, path, videos, VideoMetadata.DEFAULT_METADATA);
		return new Library(path, readOnly, videos);
	}

	private static void load(Path rootPath, Path path, List<Video> videos, VideoMetadata parentDefaultMetadata) {
		if(!Files.isDirectory(path)) {
			if(path.equals(rootPath)) {
				VideoBase.LOGGER.warn("Ignoring root path: Not a directory");
			}

			return;
		}

		if(Files.isRegularFile(path.resolve(".nomedia"))) {
			VideoBase.LOGGER.debug("Ignoring folder at " + path + ": .nomedia file exists");
			return;
		}

		LibraryMetadata libraryMetadata = null;

		VideoMetadata inferredMetadata = new VideoMetadata(new JSONObject(Map.of(
			VideoMetadata.FIELD_SERIES, path.getFileName().toString()
		)));

		VideoMetadata defaultMetadata = VideoMetadata.inherit(parentDefaultMetadata, inferredMetadata);

		Path metaPath = path.resolve("meta.json");
		if(Files.isRegularFile(metaPath)) {
			try {
				JSONObject obj = new JSONObject(Files.readString(metaPath));
				libraryMetadata = JSONConverter.decodeObject(obj, LibraryMetadata.class);
				defaultMetadata = VideoMetadata.inherit(defaultMetadata, libraryMetadata.getDefault());
			}catch(IOException | JSONParseException | ClassCastException e) {
				VideoBase.LOGGER.warn("Failed to parse metadata at " + path, e);
			}
		}

		try {
			for(Path subPath : Files.walk(path, 1, FileVisitOption.FOLLOW_LINKS)
					.filter(p -> path.equals(p.getParent()))
					.sorted(Comparator.comparing(p -> p.getFileName().toString()))
					.collect(Collectors.toList())) {
				if(Files.isDirectory(subPath)) {
					if(libraryMetadata != null && libraryMetadata.getExcludePaths().contains(subPath.getFileName().toString())) {
						VideoBase.LOGGER.debug("Ignoring folder at " + subPath + ": Listed in excludedPaths");
						continue;
					}

					load(rootPath, subPath, videos, defaultMetadata);
					continue;
				}

				String fileName = subPath.getFileName().toString();
				if(fileName.equals(METADATA_NAME) || fileName.endsWith(VIDEO_METADATA_SUFFIX)) {
					continue;
				}

				if(!VideoBase.getConfig().getIncludeFileTypes().stream().anyMatch(ext -> fileName.endsWith("." + ext))) {
					VideoBase.LOGGER.debug("Ignoring file at " + subPath + ": No valid file extension");
					continue;
				}

				String fileNameNoExt = fileName;
				if(fileNameNoExt.contains(".")) {
					fileNameNoExt = fileNameNoExt.substring(0, fileNameNoExt.lastIndexOf('.'));
				}

				VideoMetadata inferredVideoMetadata = new VideoMetadata(new JSONObject(Map.of(
					VideoMetadata.FIELD_TITLE, fileNameNoExt
				)));

				VideoMetadata videoMetadata = VideoMetadata.inherit(defaultMetadata, inferredVideoMetadata);

				Path videoMetaPath = path.resolve(subPath.getFileName().toString() + VIDEO_METADATA_SUFFIX);
				if(Files.isRegularFile(videoMetaPath)) {
					try {
						JSONObject obj = new JSONObject(Files.readString(videoMetaPath));
						videoMetadata = VideoMetadata.inherit(videoMetadata, JSONConverter.decodeObject(obj, VideoMetadata.class));
					}catch(IOException | JSONParseException | ClassCastException e) {
						VideoBase.LOGGER.warn("Failed to parse metadata at " + path, e);
					}
				}else if(libraryMetadata != null && libraryMetadata.getOverrides().containsKey(fileName)) {
					videoMetadata = VideoMetadata.inherit(videoMetadata, libraryMetadata.getOverrides().get(fileName));
				}

				String id = Hash.hash(rootPath.relativize(subPath).toString());
				videos.add(new Video(subPath, id, videoMetadata == null ? VideoMetadata.DEFAULT_METADATA : videoMetadata));
			}
		}catch(IOException e) {
			VideoBase.LOGGER.warn("Failed to load folder at " + path, e);
		}
	}

}
