package me.mrletsplay.videobase.rest;

import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import javax.imageio.ImageIO;

import me.mrletsplay.mrcore.json.JSONArray;
import me.mrletsplay.mrcore.json.JSONObject;
import me.mrletsplay.simplehttpserver.http.HttpRequestMethod;
import me.mrletsplay.simplehttpserver.http.HttpStatusCodes;
import me.mrletsplay.simplehttpserver.http.endpoint.Endpoint;
import me.mrletsplay.simplehttpserver.http.endpoint.EndpointCollection;
import me.mrletsplay.simplehttpserver.http.endpoint.RequestParameter;
import me.mrletsplay.simplehttpserver.http.header.DefaultClientContentTypes;
import me.mrletsplay.simplehttpserver.http.request.HttpRequestContext;
import me.mrletsplay.simplehttpserver.http.response.JsonResponse;
import me.mrletsplay.simplehttpserver.http.response.TextResponse;
import me.mrletsplay.simplehttpserver.http.util.MimeType;
import me.mrletsplay.simplehttpserver.http.validation.result.ValidationResult;
import me.mrletsplay.videobase.VideoBase;
import me.mrletsplay.videobase.library.Library;
import me.mrletsplay.videobase.library.Video;
import me.mrletsplay.videobase.library.VideoMetadata;
import me.mrletsplay.videobase.util.ThumbnailCreator;

public class LibraryAPI implements EndpointCollection {

	@Endpoint(method = HttpRequestMethod.GET, path = "")
	public void getLibrary(HttpRequestContext ctx) {
		String groupBy = ctx.getRequestedPath().getQuery().getFirst("groupBy", "none").toLowerCase();

		Library library = VideoBase.getLibrary();

		JSONObject libraryObj = new JSONObject();

		JSONObject videosObj = new JSONObject();

		List<String> groups;
		switch(groupBy) {
			case "series": {
				groups = library.getSeries();
				break;
			}
			case "author": {
				groups = library.getAuthors();
				break;
			}
			case "none":
			default:
				groups = List.of("all");
				break;
		}

		for(String group : groups) {
			List<Video> videos;
			switch(groupBy) {
				case "series": {
					videos = library.findVideosBy(group, null);
					break;
				}
				case "author": {
					videos = library.findVideosBy(null, group);
					break;
				}
				case "none":
				default:
					videos = library.getVideos();
					break;
			}

			JSONArray videosArr = new JSONArray();
			for(Video video : videos) {
				videosArr.add(videoToJSON(video));
			}
			videosObj.put(group, videosArr);
		}

		libraryObj.put("videos", videosObj);

		ctx.respond(HttpStatusCodes.OK_200, new JsonResponse(libraryObj));
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/video/{video}", pathPattern = true)
	public void getVideo(HttpRequestContext ctx, @RequestParameter("video") String videoId) {
		Video video = VideoBase.getLibrary().findVideoById(videoId);
		if(video == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Not found"));
			return;
		}

		ctx.respond(HttpStatusCodes.OK_200, new JsonResponse(videoToJSON(video)));
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/video/{video}/metadata", pathPattern = true)
	public void getVideoMetadata(HttpRequestContext ctx, @RequestParameter("video") String videoId) {
		Video video = VideoBase.getLibrary().findVideoById(videoId);
		if(video == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Not found"));
			return;
		}

		VideoMetadata videoMetadata = video.getMetadata();

		JSONObject meta = new JSONObject();
		meta.put("raw", videoMetadata.getRaw());
		meta.put("inherited", videoMetadata.getParent() == null ? new JSONObject() : videoMetadata.getParent().getEffective());
		meta.put("effective", videoMetadata.getEffective());

		ctx.respond(HttpStatusCodes.OK_200, new JsonResponse(meta));
	}

	@Endpoint(method = HttpRequestMethod.PUT, path = "/video/{video}/metadata", pathPattern = true)
	public void updateVideoMetadata(HttpRequestContext ctx, @RequestParameter("video") String videoId) {
		Library library = VideoBase.getLibrary();
		if(library.isReadOnly()) {
			ctx.respond(HttpStatusCodes.ACCESS_DENIED_403, new TextResponse("Library is read-only"));
			return;
		}

		Video video = VideoBase.getLibrary().findVideoById(videoId);
		if(video == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Not found"));
			return;
		}

		JSONObject metadata = ctx.expectContent(DefaultClientContentTypes.JSON_OBJECT);
		if(metadata == null) {
			ctx.respond(HttpStatusCodes.BAD_REQUEST_400, ValidationResult.error("_", "Invalid content").asJsonResponse());
			return;
		}

		ValidationResult result = VideoMetadata.VALIDATOR.validate(metadata);
		if(!result.isOk()) {
			ctx.respond(HttpStatusCodes.BAD_REQUEST_400, result.asJsonResponse());
			return;
		}

		if(!library.updateMetadata(video, new VideoMetadata(metadata))) {
			ctx.respond(HttpStatusCodes.INTERNAL_SERVER_ERROR_500, new TextResponse("Failed to update metadata"));
			return;
		}

		ctx.respond(HttpStatusCodes.OK_200, new JsonResponse(video.getMetadata().getRaw()));
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/video/{video}/thumbnail", pathPattern = true)
	public void getThumbnail(HttpRequestContext ctx, @RequestParameter("video") String videoId) {
		Video video = VideoBase.getLibrary().findVideoById(videoId);
		if(video == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Not found"));
			return;
		}

		BufferedImage thumbnail = ThumbnailCreator.getThumbnail(video);
		if(thumbnail == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("No thumbnail found"));
			return;
		}

		ByteArrayOutputStream bOut = new ByteArrayOutputStream();
		try {
			ImageIO.write(thumbnail, "PNG", bOut);
		} catch (IOException e) {
			ctx.respond(HttpStatusCodes.INTERNAL_SERVER_ERROR_500, new TextResponse("Failed to create thumbnail"));
		}

		ctx.getServerHeader().setContent(MimeType.PNG, bOut.toByteArray());
	}

	@Endpoint(method = HttpRequestMethod.GET, path = "/video/{video}/stream", pathPattern = true)
	public void getStream(HttpRequestContext ctx, @RequestParameter("video") String videoId) {
		Video video = VideoBase.getLibrary().findVideoById(videoId);
		if(video == null) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Not found"));
			return;
		}

		Path videoPath = video.getPath();
		if(!Files.isRegularFile(videoPath)) {
			ctx.respond(HttpStatusCodes.NOT_FOUND_404, new TextResponse("Video file not found"));
			return;
		}

		try {
			long length = Files.size(videoPath);
			String contentType = Files.probeContentType(videoPath);
			InputStream in = Files.newInputStream(videoPath);
			ctx.getServerHeader().setCompressionEnabled(false);
			ctx.getServerHeader().setContent(MimeType.of(contentType == null ? "video/mpeg" : contentType), in, length);
		} catch (IOException e) {
			ctx.respond(HttpStatusCodes.INTERNAL_SERVER_ERROR_500, new TextResponse("Failed to load video"));
			return;
		}
	}

	private JSONObject videoToJSON(Video video) {
		JSONObject v = new JSONObject();
		v.put("id", video.getId());
		v.put("metadata", video.getMetadata().getEffective());
		return v;
	}

	@Override
	public String getBasePath() {
		return "/api/library";
	}

}
