package me.mrletsplay.videobase.util;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.imageio.ImageIO;

import me.mrletsplay.videobase.VideoBase;
import me.mrletsplay.videobase.library.Video;

public class ThumbnailCreator {

	private static Map<String, BufferedImage> cachedThumbnails = new ConcurrentHashMap<>();
	private static ExecutorService executorService = Executors.newFixedThreadPool(10);

	public static void clearCache() {
		cachedThumbnails.clear();
	}

	public static void createThumbnails() {
		for(Video v : VideoBase.getLibrary().getVideos()) {
			executorService.submit(() -> createThumbnail(v));
		}
	}

	public static BufferedImage createThumbnail(Video video) {
		if(cachedThumbnails.containsKey(video.getId())) {
			return cachedThumbnails.get(video.getId());
		}

		String cachePath = VideoBase.getConfig().getCachePath();
		Path thumbnailFile = cachePath == null ? null : Path.of(cachePath).resolve("thumbnails").resolve(video.getId() + ".png").toAbsolutePath();

		if(thumbnailFile != null && Files.isRegularFile(thumbnailFile)) {
			try(InputStream in = Files.newInputStream(thumbnailFile)) {
				BufferedImage thumbnail = ImageIO.read(in);
				if(thumbnail == null) return null;

				cachedThumbnails.put(video.getId(), thumbnail);
				return thumbnail;
			} catch (IOException e) {
				VideoBase.LOGGER.warn("Failed to load thumbnail", e);
				return null;
			}
		}

		ProcessBuilder ffmpegBuilder = new ProcessBuilder("ffmpeg",
			"-i", video.getPath().toAbsolutePath().toString(),
			"-r", "0.5",
			"-vf", "thumbnail,scale=960:540:force_original_aspect_ratio=decrease,pad=960:540:-1:-1:color=black",
			"-frames:v", "1",
			"-f", "image2pipe",
			"-vcodec", "png",
			"-"
		);

		try {
			Process ffmpeg = ffmpegBuilder.start();
			long start = System.currentTimeMillis();

			ByteArrayOutputStream bOut = new ByteArrayOutputStream();
			byte[] buf = new byte[1024];
			int len;
			while(System.currentTimeMillis() - start < 60000
				&& (len = ffmpeg.getInputStream().read(buf)) != -1) {
				bOut.write(buf, 0, len);
			}

			ffmpeg.waitFor(60, TimeUnit.SECONDS);
			if(ffmpeg.isAlive()) {
				VideoBase.LOGGER.warn("Video " + video.getId() + ": ffmpeg didn't exit after 60 seconds, destroying");
				ffmpeg.destroy();

				if(!ffmpeg.waitFor(60, TimeUnit.SECONDS)) {
					VideoBase.LOGGER.warn("Video " + video.getId() + ": ffmpeg didn't exit 60 seconds after destroying, destroying forcibly");
					ffmpeg.destroyForcibly();
				}

				return null;
			}

			BufferedImage thumbnail = ImageIO.read(new ByteArrayInputStream(bOut.toByteArray()));
			if(thumbnail == null) return null;

			if(thumbnailFile != null) {
				Files.createDirectories(thumbnailFile.getParent());
				try(OutputStream out = Files.newOutputStream(thumbnailFile)) {
					ImageIO.write(thumbnail, "PNG", out);
				}
			}

			VideoBase.LOGGER.debug("Created thumbnail for video at " + video.getPath());
			cachedThumbnails.put(video.getId(), thumbnail);
			return thumbnail;
		} catch (InterruptedException | IOException e) {
			VideoBase.LOGGER.warn("Failed to create thumbnail", e);
			return null;
		}
	}

	public static BufferedImage getThumbnail(Video video) {
		return cachedThumbnails.get(video.getId());
	}

}
